/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

#ifndef _EMU_INC_CIPHER_AES_H
#define _EMU_INC_CIPHER_AES_H

#include "emu_general.h"

#include "inc_vendor.h"
#include "inc_cipher_aes.h"

extern u32a te0[256];
extern u32a te1[256];
extern u32a te2[256];
extern u32a te3[256];
extern u32a te4[256];

extern u32a td0[256];
extern u32a td1[256];
extern u32a td2[256];
extern u32a td3[256];
extern u32a td4[256];

#endif // _EMU_INC_CIPHER_AES_H
