/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

#include "types.h"
#include "common.h"
#include "emu_general.h"

#include "inc_vendor.h"
#include "inc_platform.h"
#include "inc_ecc_secp256k1.cl"

