/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

#ifndef _EMU_INC_HASH_SHA256_H
#define _EMU_INC_HASH_SHA256_H

#include "emu_general.h"

#include "inc_vendor.h"
#include "inc_hash_sha256.h"

extern u32a k_sha256[64];

#endif // _EMU_INC_HASH_SHA256_H
