#include <windows.h>
#include <commctrl.h>

VS_VERSION_INFO VERSIONINFO
FILEVERSION 6, 1, 100, 3756
PRODUCTVERSION 6, 1, 100, 3756
FILEOS VOS__WINDOWS32
FILETYPE VFT_APP
{
  BLOCK "StringFileInfo"
  {
    BLOCK "040904E4"
    {
      VALUE "CompanyName", "Alexander Roshal\0"
      VALUE "ProductName", "RAR decompression library\0"
      VALUE "FileDescription", "RAR decompression library\0"
      VALUE "FileVersion", "6.1.0\0"
      VALUE "ProductVersion", "6.1.0\0"
      VALUE "LegalCopyright", "Copyright © Alexander Roshal 1993-2021\0"
      VALUE "OriginalFilename", "Unrar.dll\0"
    }
  }
  BLOCK "VarFileInfo"
  {
    VALUE "Translation", 0x0409, 0x04E4
  }
}

