/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

#ifdef KERNEL_STATIC
#include "inc_vendor.h"
#include "inc_types.h"
#include "inc_platform.cl"
#include "inc_common.cl"
#endif

KERNEL_FQ void m02000_m04 (KERN_ATTR_BASIC ())
{
}

KERNEL_FQ void m02000_m08 (KERN_ATTR_BASIC ())
{
}

KERNEL_FQ void m02000_m16 (KERN_ATTR_BASIC ())
{
}

KERNEL_FQ void m02000_s04 (KERN_ATTR_BASIC ())
{
}

KERNEL_FQ void m02000_s08 (KERN_ATTR_BASIC ())
{
}

KERNEL_FQ void m02000_s16 (KERN_ATTR_BASIC ())
{
}
